import inspect

from pkg_resources import resource_filename as pkg_resource_filename
from pkg_resources import resource_stream as pkg_resource_stream


def external_caller():
    i = inspect.stack()
    names = (inspect.getmodule(i[x][0]).__name__ for x in range(len(i)))
    return next(name for name in names if name != __name__)


def resource_path(subpath):
    module_name = external_caller()
    return pkg_resource_filename(module_name, subpath)


def resource_stream(subpath):
    module_name = external_caller()
    return pkg_resource_stream(module_name, subpath)


def resource_data(subpath):
    return resource_stream(subpath).read()


def resource_text(subpath, encoding="utf-8"):
    return resource_stream(subpath).read().decode(encoding)



from pathlib import Path


def save_xlsx(sheet_mapping, destination):
    destination = Path(destination).expanduser()
    from xlsxwriter import Workbook

    workbook = Workbook(str(destination))

    for name, rows in sheet_mapping.items():
        worksheet = workbook.add_worksheet(name)

        for r, row in enumerate([rows.keys()] + rows):
            for c, col in enumerate(row):
                worksheet.write(r, c, col)

    workbook.close()



import string

from .fileutil import is_mapping
from .resultset import resultproxy_to_results

INSERT = """
    insert into
        {table} ({colspec})
    values
        ({valuespec})
"""


INSERT_DEFAULT = """
    insert into
        {table}
    default values
"""


INSERT_UPSERT = """
    on conflict ({upsertkeyspec})
    do update set
        {upsertspec}
"""


INSERT_UPSERT = """
    on conflict ({upsertkeyspec})
    do update set
        {upsertspec}

"""


INSERT_UPSERT_DO_NOTHING = """
    on conflict ({upsertkeyspec})
    do nothing
"""


VALID = {k: None for k in (string.ascii_letters + string.digits + "_")}


def valid(key):
    clist = [_ for _ in list(key) if _ in VALID]
    v = "".join(clist)
    if v != key:
        raise ValueError(f'invalid key: "{key}"')
    return v


def insert(s, table, rows, upsert_on=None, returning=None):

    if is_mapping(rows):
        rows = [rows]

    if not rows:
        raise ValueError("empty list of rows, nothing to upsert")

    if returning is None:
        returning = not len(rows) > 1

    keys = [valid(k) for k in rows[0].keys()]

    if keys:
        colspec = ", ".join([f'"{k}"' for k in keys])
        valuespec = ", ".join(":{}".format(k) for k in keys)
        q = INSERT.format(table=table, colspec=colspec, valuespec=valuespec)
    else:
        q = INSERT_DEFAULT.format(table=table)

    if upsert_on:
        upsert_keys = list(keys)

        for k in upsert_on:
            upsert_keys.remove(k)

        upsertkeyspec = ", ".join([f'"{k}"' for k in upsert_on])

        if upsert_keys:
            upsertspec = ", ".join(f'"{k}" = excluded."{k}"' for k in upsert_keys)

            q_upsert = INSERT_UPSERT.format(
                upsertkeyspec=upsertkeyspec, upsertspec=upsertspec
            )
        else:
            q_upsert = INSERT_UPSERT_DO_NOTHING.format(upsertkeyspec=upsertkeyspec)

        q = q + q_upsert
    if returning:
        q += " returning *"

    rp = s.execute(q, rows)
    return resultproxy_to_results(rp)



from contextlib import contextmanager
from functools import partial
from pathlib import Path

from schemainspect import get_inspector
from sqlbag import S, raw_execute

from .inserting import insert
from .migration import SchemaManagement
from .paging import Paging, paging_wrapped_query
from .result import Result
from .resultset import resultproxy_to_results


def build_proc_call_query(_proc_name, *args, **kwargs):
    _proc_name = _proc_name.replace("__", ".")
    params = {f"positional{i}": x for i, x in enumerate(args)}
    params.update(**dict(kwargs))
    paramnames = params.keys()
    bindparams = [f":{name}" for name in list(paramnames)]
    paramspec = ", ".join(bindparams)
    query = f"select * from {_proc_name}({paramspec})"
    return query, params


class transactionprocs:
    def __init__(self, sess):
        self.session = sess

    def __getattr__(self, name):
        return partial(self.call, name)

    def call(self, _proc_name, *args, **kwargs):
        query, params = build_proc_call_query(_proc_name, *args, **kwargs)

        return self.session.ex(query, params)


EXPLAIN_DEFAULTS = dict(
    analyze=False,
    verbose=False,
    costs=True,
    buffers=False,
    timing=True,
    format="TEXT",
    summary=None,
)


def explain_prefix(**kwargs):

    # summary on by default if analyze is on
    defaults = dict(EXPLAIN_DEFAULTS)
    defaults["summary"] = defaults["analyze"]

    flags = {
        k: kwargs.get(k)
        for k, default in defaults.items()
        if kwargs.get(k) and kwargs.get(k) != default
    }

    def make_flag(k, v):
        if isinstance(v, bool):
            return k.upper()
        return f"{k} {v}".upper()

    flags = ", ".join(make_flag(k, v) for k, v in sorted(flags.items()))

    prefix = "EXPLAIN"
    if flags:
        prefix += f" ({flags})"

    return prefix


class transaction:
    def __init__(self, s):
        self.s = s
        self.procs = transactionprocs(self)

    def ex(self, *args, execution_options=None, **kwargs):
        if execution_options:
            _resultproxy = self.s.connection(
                execution_options=execution_options
            ).execute(*args, **kwargs)
        else:
            _resultproxy = self.s.execute(*args, **kwargs)
        results = resultproxy_to_results(_resultproxy)
        return results

    def raw_ex(self, *args, execution_options=None, **kwargs):
        if execution_options:
            return self.s.connection(execution_options=execution_options).execute(
                *args, **kwargs
            )
        else:
            return self.s.execute(*args, **kwargs)

    def paged(
        self,
        query,
        *args,
        bookmark=None,
        ordering,
        per_page,
        backwards,
        use_top=False,
        supports_row=True,
        **kwargs,
    ):
        query, page_params = paging_wrapped_query(
            query, ordering, bookmark, per_page, backwards, use_top, supports_row
        )

        argslist = list(args)
        try:
            params = argslist[0]
        except IndexError:
            argslist.append({})
            params = argslist[0]

        params.update(page_params)

        argslist[0] = params
        args = tuple(argslist)

        results = self.ex(query, *args, **kwargs)
        results.paging = Paging(results, per_page, ordering, bookmark, backwards)

        return results

    def insert(self, table, rowlist, upsert_on=None, returning=None):
        return insert(self.s, table, rowlist, upsert_on, returning)

    def pg_notify(self, channel, payload=None):
        return self.ex(
            """
            select pg_notify(:channel, :payload)
        """,
            dict(channel=channel, payload=payload),
        )

    def explain(self, *args, **kwargs):
        args = list(args)

        stmt = args[0]

        prefix = explain_prefix(**kwargs)

        for k in EXPLAIN_DEFAULTS:
            kwargs.pop(k, None)

        args[0] = f"{prefix} {stmt}"
        return self.ex(*args, **kwargs)


class procs:
    def __init__(self, db):
        self.db = db

    def __getattr__(self, name):
        return partial(self.call, name)

    def call(self, _proc_name, *args, **kwargs):
        _proc_name = _proc_name.replace("__", ".")

        with self.db.transaction() as s:
            query, params = build_proc_call_query(_proc_name, *args, **kwargs)
            return s.ex(query, params)


class db(SchemaManagement):
    def __init__(self, *args, **kwargs):
        self._args = args
        if args and args[0].startswith("postgres"):
            kwargs.setdefault("executemany_mode", "batch")
        self._kwargs = kwargs

    @contextmanager
    def transaction(self):
        with S(*self._args, **self._kwargs) as s:
            yield transaction(s)

    @property
    def procs(self):
        return procs(self)

    @property
    def db_url(self):
        return self._args[0]

    def ss(self, *args, **kwargs):
        with self.transaction() as t:
            return t.ex(*args, **kwargs)

    def ss_iter(self, *args, **kwargs):
        with self.transaction() as t:
            for row in t.raw_ex(*args, **kwargs):
                yield Result(row)

    def paged(self, *args, **kwargs):
        with self.transaction() as t:
            return t.paged(*args, **kwargs)

    def raw_from_file(self, f):
        sql = Path(f).read_text()
        return self.raw(sql)

    def ss_from_file(self, f, *args, **kwargs):
        sql = Path(f).read_text()
        return self.ss(sql, *args, **kwargs)

    def paged_from_file(self, f, *args, **kwargs):
        sql = Path(f).read_text()
        return self.paged(sql, *args, **kwargs)

    def insert(self, table, rowlist, upsert_on=None, returning=None):
        with self.transaction() as s:
            inserted = s.insert(table, rowlist, upsert_on, returning)
        return inserted

    def raw(self, sql):
        with S(*self._args, **self._kwargs) as s:
            _results = raw_execute(s, sql)
            return _results

    def inspect(self):
        with S(*self._args, **self._kwargs) as s:
            i = get_inspector(s)
        return i

    def pg_notify(self, channel, payload=None):
        with self.transaction() as t:
            return t.pg_notify(channel, payload)

    def explain(self, *args, **kwargs):
        with self.transaction() as t:
            return t.explain(*args, **kwargs)



from logx import log
from sqlbag import create_database  # noqa
from sqlbag.pg import use_pendulum_for_time_types  # noqa

from .cleaning import standardize_key, standardized_key_mapping  # noqa
from .connections import db  # noqa
from .fileutil import file_text, files, from_file, from_files  # noqa
from .openers import (  # noqa
    csv_column_names,
    csv_rows_it,
    detect_enc,
    detect_string_enc,
    dicts_from_rows,
    first_n_lines,
    from_csv,
    from_psv,
    from_tsv,
    from_xls,
    from_xlsx,
    smart_open,
    sniff_csv_dialect,
)
from .resources import (  # noqa
    resource_data,
    resource_path,
    resource_stream,
    resource_text,
)
from .result import Result  # noqa
from .resultset import Results  # noqa
from .saving import save_xlsx  # noqa
from .sqlutil import create_table_statement  # noqa
from .typeguess import (  # noqa
    guess_column_type,
    guess_sql_column_type,
    guess_value_type,
)
from .util import noneslarger, nonessmaller, trunc  # noqa
from .uuids import deterministic_uuid  # noqa

log.set_null_handler()
use_pendulum_for_time_types()



from pathlib import Path

from py._path.local import LocalPath


def is_pathlike(f):
    return isinstance(f, (str, Path, LocalPath))


def noneslarger(x):
    if x is None:
        return 1, x
    else:
        return 0, x


def nonessmaller(x):
    if x is None:
        return -1, x
    else:
        return 0, x


ELLIPSIS = "\u2026"


def trunc(s, length=25):
    if len(s) > length:
        s = s[:length] + ELLIPSIS
    return s



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['results', 'results.openers']

package_data = \
{'': ['*']}

install_requires = \
['chardet',
 'furl',
 'logx',
 'markdown',
 'migra',
 'pendulum',
 'psycopg2-binary',
 'sqlalchemy>=1.3',
 'sqlbag>=0.1.1548994599',
 'tabulate']

extras_require = \
{'excel': ['xlrd', 'openpyxl<2.6.3', 'xlsxwriter']}

setup_kwargs = {
    'name': 'results',
    'version': '0.1.1585546315',
    'description': "Don't get mad, get results",
    'long_description': '# Don\'t get mad, get `results`\n\nTabular data and SQL for people who don\'t have time to faff about.\n\nMove between xlsx, xls, csv, python, postgres and back with ease.\n\nFeatures:\n\n- Zero-boilerplate database creating, connecting and querying.\n- Loading/tidying/transforming csv and excel data.\n- Autodetect column types, load your data with little or no manual specification.\n- Powerful multi-column, multi-order keyset paging of database results.\n- Schema syncing.\n\nLimitations\n\n- Python 3.6+, PostgreSQL 10+ only. Many features will work with other databases, but many won\'t. Just use Postgres!\n\n# Installation\n\n[`results` is on PyPI](https://pypi.org/project/results). Install it with `pip` or any of the (many) Python package managers.\n\n## Scenario\n\nSomebody gives you a messy csv or excel file. You need to load it, clean it up, put it into a database, query it, make a pivot table from it, then send the pivot table to somebody as a csv.\n\n`results` is here to get this sort of thing done quickly and with minimum possible fuss.\n\nLet\'s see.\n\nFirst, load and clean:\n\n```python\nimport results\n\n# load a csv (in this example, some airport data)\nsheet = results.from_file("airports.csv")\n\n# do general cleanup\nsheet.standardize_spaces()\nsheet.set_blanks_to_none()\n\n# give the keys lowercase-with-underscore names to keep the database happy\ncleaned = sheet.with_standardized_keys()\n```\n\nThen, create a database:\n\n```python\n# create a database\nDB = "postgresql:///example"\n\ndb = results.db(DB)\n\n# create it if it doesn\'t exist\ndb.create_database()\n```\n\nThen create a table for the data, automatically guessing the columns and creating a table to match.\n\n```python\n# guess the column types\nguessed = cleaned.guessed_sql_column_types()\n\n# create a table for the data\ncreate_table_statement = results.create_table_statement("data", guessed)\n\n# create or auto-update the table structure in the database\n# syncing requires a copy of postgres running locally with your current user set up as superuser\ndb.sync_db_structure_to_definition(create_table_statement, confirm=False)\n```\n\nThen insert the data and freely query it.\n\n```python\n# insert the data. you can also do upserts with upsert_on!\ndb.insert("data", cleaned)\n\n# show recent airfreight numbers from the top 5 airports\n# ss means "single statement"\nquery_result = db.ss(\n    """\nwith top5 as (\n    select\n        foreignport, sum(freight_in_tonnes)\n    from\n        data\n    where year >= 2010\n    group by\n        foreignport\n    order by 2 desc\n    limit 5\n)\n\nselect\n    year, foreignport, sum(freight_in_tonnes)\nfrom\n    data\nwhere\n    year >= 2010\n    and foreignport in (select foreignport from top5)\ngroup by 1, 2\norder by 1, 2\n\n"""\n)\n```\n\nCreate a pivot table, then print it as markdown or save it as csv.\n\n```python\n# create a pivot table\npivot = query_result.pivoted()\n\n# print the pivot table in markdown format\nprint(pivot.md)\n```\n\nOutput:\n\n```\n|   year |   Auckland |    Dubai |   Hong Kong |   Kuala Lumpur |   Singapore |\n|-------:|-----------:|---------:|------------:|---------------:|------------:|\n|   2010 |     288997 | 145527   |      404735 |       226787   |      529407 |\n|   2011 |     304628 | 169868   |      428990 |       244053   |      583921 |\n|   2012 |     312828 | 259444   |      400596 |       272093   |      614155 |\n|   2013 |     306783 | 257263   |      353895 |       272804   |      592886 |\n|   2014 |     309318 | 244776   |      330521 |       261438   |      620419 |\n|   2015 |     286202 | 263378   |      290292 |       252906   |      633862 |\n|   2016 |     285973 | 236419   |      309556 |       175858   |      614172 |\n|   2017 |     314405 | 226048   |      340216 |       199868   |      662505 |\n|   2018 |     126712 |  91611.2 |      134540 |        74667.5 |      250653 |\n```\n\nSave the table as a csv:\n\n```python\npivot.save_csv("2010s_freight_sources_top5.csv")\n```\n\n## Design philosophy\n\n- Avoid boilerplate at all costs. Make it as simple as possible but no simpler.\n\n- Don\'t reinvent the wheel: `results` uses sqlalchemy for database connections, existing excel parsing libraries for excel parsing, etc etc. `results` brings it all together, sprinkles some sugar on top, and puts it at your fingertips.\n\n- Eat your own dogfood: We use this ourselves every day.\n\n## Documentation\n\nThis README.md is currently all there is :( But we\'ll add more soon, we promise!\n\n## Credits\n\n- [Rob](https://github.com/djrobstep)\n- [Jason](https://github.com/jasongi)\n- [Nick](https://github.com/nmcl23)\n\n## Contributions\n\nYes please!\n',
    'author': 'Robert Lechte',
    'author_email': 'rlechte@actu.org.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import pendulum
import six
from pendulum.date import Date as date_type
from pendulum.datetime import DateTime as datetime_type


def guess_value_type(x):
    if x == "":
        return None

    try:
        f = float(x)
        if f.is_integer():
            return int
        else:
            return float
    except (ValueError, TypeError):
        pass

    try:
        p = pendulum.parse(x, exact=True)

        if isinstance(p, datetime_type):
            return datetime_type

        if isinstance(p, date_type):
            return date_type

    except Exception:
        pass

    return six.text_type


def guess_column_type(values):
    present = {guess_value_type(x) for x in values if x is not None}

    if six.text_type in present:
        return six.text_type
    if datetime_type in present:
        return datetime_type
    if date_type in present:
        return date_type
    if float in present:
        return float
    if int in present:
        return int
    return six.text_type


def guess_sql_column_type(values):
    return PY_TO_SQL[guess_column_type(values)]


PY_TO_SQL = {
    six.text_type: "text",
    datetime_type: "timestamp",
    date_type: "date",
    float: "decimal",
    int: "int",
    None: "text",
}



ASC_OR_DESC = ("desc", "asc")


def quoted(column_name):
    return f'"{column_name}"'


def parse_ordering_col(c):
    tokens = c.rsplit(None, 1)

    if len(tokens) > 1:
        possible_direction = tokens[-1].lower()
        if possible_direction in ASC_OR_DESC:
            c = tokens[0]
            descending = possible_direction == "desc"
        else:
            descending = False
    else:
        descending = False

    c = c.strip().strip('"')
    return c, descending


def parse_ordering(ordering):
    cols = ordering.split(",")
    return [parse_ordering_col(c) for c in cols]


def reversed_ordering(ordering):
    return [(c[0], not c[1]) for c in ordering]


def ordering_from_parsed(cols):
    clist = [(quoted(c) + (descending and " desc" or "")) for c, descending in cols]

    return ", ".join(clist)


def create_table_statement(table_name, colspec):
    quoted_name = quoted(table_name)
    colspec = ",\n  ".join(f"{quoted(k)} {v}" for k, v in colspec.items())

    return f"""create table {quoted_name} (
  {colspec}
);

"""
